#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import pandas as pd
from covsirphy.util.term import Term


class _RemoteDatabase(Term):
    """
    Base class to access remote databases.

    Args:
        filename (str): CSV filename to save records
    """
    # Citation
    CITATION = ""
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    COL_DICT = {}

    def __init__(self, filename):
        # Filepath to save files
        try:
            self.filepath = Path(filename)
        except TypeError:
            raise TypeError(f"@filename should be a path-like object, but {filename} was applied.")
        self.filepath.parent.mkdir(exist_ok=True, parents=True)
        # List of primary sources
        self.primary_list = []
        # List of column names (defined in Term class)
        self.saved_cols = list(self.COL_DICT.values())

    def to_dataframe(self, force, verbose):
        """
        Load the dataset and return it as a dataframe.

        Args:
            force (bool): whether download the dataset from server or not when we have saved dataset
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()
        """
        # Read local file if available and usable
        if not force and self.filepath.exists():
            try:
                return self._ensure_dataframe(self.read(), columns=self.saved_cols)
            except ValueError:
                # ValueError: Usecols do not match columns
                pass
        # Download dataset from server
        df = self.download(verbose=verbose)
        df = df.rename(columns=self.COL_DICT)
        df = self._ensure_dataframe(df, columns=self.saved_cols)
        df.to_csv(self.filepath, index=False)
        return df

    def read(self):
        """
        Load the dataset with a local file.

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by .COL_DICT.values()
        """
        kwargs = {"low_memory": False, "header": 0, "usecols": self.saved_cols}
        return pd.read_csv(self.filepath, **kwargs)

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by .COL_DICT.values()
        """
        raise NotImplementedError

    @property
    def primary(self):
        """
        str: the list of primary sources.
        """
        return "\n".join(self.primary_list)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['covsirphy',
 'covsirphy.analysis',
 'covsirphy.automl',
 'covsirphy.cleaning',
 'covsirphy.loading',
 'covsirphy.ode',
 'covsirphy.phase',
 'covsirphy.regression',
 'covsirphy.simulation',
 'covsirphy.trend',
 'covsirphy.util',
 'covsirphy.visualization',
 'covsirphy.worldwide']

package_data = \
{'': ['*']}

install_requires = \
['AutoTS>=0.4.0,<0.5.0',
 'Unidecode>=1.2.0,<2.0.0',
 'better-exceptions>=0.3.2,<0.4.0',
 'country-converter>=0.7.1,<0.8.0',
 'covid19dh>=2.0.3,<3.0.0',
 'descartes>=1.1.0,<2.0.0',
 'funcparserlib==1.0.0a0',
 'geopandas>=0.9,<0.11',
 'japanmap>=0.0.21,<0.0.25',
 'lightgbm>=3.2.1,<4.0.0',
 'mapclassify>=2.4.2,<3.0.0',
 'matplotlib>=3.2.1,<4.0.0',
 'numpy>=1.18.5,<2.0.0',
 'optuna>=2.3.0,<3.0.0',
 'pandas>=1.1.5,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 'ruptures>=1.1.1,<2.0.0',
 'scikit-learn>=0.24,<1.1',
 'scipy>=1.4.1,<2.0.0',
 'seaborn>=0.11.1,<0.12.0',
 'tabulate>=0.8.7,<0.9.0',
 'wbdata>=0.3.0,<0.4.0']

setup_kwargs = {
    'name': 'covsirphy',
    'version': '2.24.0',
    'description': 'COVID-19 data analysis with phase-dependent SIR-derived ODE models',
    'long_description': '|PyPI version| |Downloads| |PyPI - Python Version| |Build Status|\n|GitHub license| |Maintainability| |Test Coverage| |Open Source Helpers|\n\nCovsirPhy introduction\n======================\n\n`Documentation <https://lisphilar.github.io/covid19-sir/index.html>`__\n\\|\n`Installation <https://lisphilar.github.io/covid19-sir/INSTALLATION.html>`__\n\\| `Quickest\nusage <https://lisphilar.github.io/covid19-sir/usage_quickest.html>`__\n\\| `API\nreference <https://lisphilar.github.io/covid19-sir/covsirphy.html>`__ \\|\n`GitHub <https://github.com/lisphilar/covid19-sir>`__ \\| `Qiita\n(Japanese) <https://qiita.com/tags/covsirphy>`__\n\nCovsirPhy is a Python library for COVID-19 (Coronavirus disease 2019)\ndata analysis with phase-dependent SIR-derived ODE models. We can\ndownload datasets and analyse them easily. Scenario analysis with\nCovsirPhy enables us to make data-informed decisions.\n\nFunctionalities\n---------------\n\n-  `Data preparation and data\n   visualization <https://lisphilar.github.io/covid19-sir/usage_dataset.html>`__\n-  `Phase setting with S-R Trend\n   analysis <https://lisphilar.github.io/covid19-sir/usage_phases.html>`__\n-  `Numerical simulation of ODE\n   models <https://lisphilar.github.io/covid19-sir/usage_theoretical.html>`__:\n   SIR, SIR-D and SIR-F model\n-  `Phase-dependent parameter estimation of ODE\n   models <https://lisphilar.github.io/covid19-sir/usage_quickest.html>`__\n-  `Scenario\n   analysis <https://lisphilar.github.io/covid19-sir/usage_quick.html>`__:\n   Simulate the number of cases with user-defined parameter values\n-  `Predict the parameter valuse to forecast the number of\n   cases. <https://lisphilar.github.io/covid19-sir/usage_quick.html#Short-term-prediction-of-parameter-values>`__\n\nInspiration\n-----------\n\n-  Monitor the spread of COVID-19\n-  Keep track parameter values/reproduction number in each\n   country/province\n-  Find the relationship of reproductive number and measures taken by\n   each country\n\nIf you have ideas or need new functionalities, please join this project.\nAny suggestions with `Github\nIssues <https://github.com/lisphilar/covid19-sir/issues/new/choose>`__\nare always welcomed. Questions are also great. Please read `Guideline of\ncontribution <https://lisphilar.github.io/covid19-sir/CONTRIBUTING.html>`__\nin advance.\n\nInstallation\n------------\n\nThe latest stable version of CovsirPhy is available at `PyPI (The Python\nPackage Index): covsirphy <https://pypi.org/project/covsirphy/>`__ and\nsupports Python 3.7 or newer versions. Details are explained in\n`Documentation:\nInstallation <https://lisphilar.github.io/covid19-sir/INSTALLATION.html>`__.\n\n.. code:: bash\n\n    pip install --upgrade covsirphy\n\nUsage\n-----\n\nQuickest tour of CovsirPhy is here. The following codes analyze the\nrecords in Japan, but we can change the country name when creating\n``Scenario`` class instance for your own analysis.\n\n.. code:: python\n\n    import covsirphy as cs\n    # Download and update datasets\n    data_loader = cs.DataLoader("input")\n    jhu_data = data_loader.jhu()\n    # Select country name and register the data\n    snl = cs.Scenario(country="Japan")\n    snl.register(jhu_data)\n    # Check records\n    snl.records()\n    # S-R trend analysis\n    snl.trend().summary()\n    # Parameter estimation of SIR-F model\n    snl.estimate(cs.SIRF)\n    # History of reproduction number\n    _ = snl.history(target="Rt")\n    # History of parameters\n    _ = snl.history_rate()\n    _ = snl.history(target="rho")\n    # Simulation for 30 days\n    snl.add(days=30)\n    _ = snl.simulate()\n\nFurther information:\n\n-  `CovsirPhy\n   documentation <https://lisphilar.github.io/covid19-sir/index.html>`__\n-  `Kaggle: COVID-19 data with SIR\n   model <https://www.kaggle.com/lisphilar/covid-19-data-with-sir-model>`__\n\nRelease notes\n-------------\n\nRelease notes are\n`here <https://github.com/lisphilar/covid19-sir/releases>`__. Titles &\nlinks of issues are listed with acknowledgement.\n\nWe can see the release plan for the next stable version in `milestone\npage of the GitHub\nrepository <https://github.com/lisphilar/covid19-sir/milestones>`__. If\nyou find a highly urgent matter, please let us know via `issue\npage <https://github.com/lisphilar/covid19-sir/issues>`__.\n\nSupport\n-------\n\nPlease support this project as a developer (or a backer). |Become a\nbacker|\n\nDevelopers\n----------\n\nCovsirPhy library is developed by a community of volunteers. Please see\nthe full list\n`here <https://github.com/lisphilar/covid19-sir/graphs/contributors>`__.\n\nThis project started in Kaggle platform. Lisphilar published `Kaggle\nNotebook: COVID-19 data with SIR\nmodel <https://www.kaggle.com/lisphilar/covid-19-data-with-sir-model>`__\non 12Feb2020 and developed it, discussing with Kaggle community. On\n07May2020, "covid19-sir" repository was created. On 10May2020,\n``covsirphy`` version 1.0.0 was published in GitHub. First release in\nPyPI (version 2.3.0) was on 28Jun2020.\n\nLicense: Apache License 2.0\n---------------------------\n\nPlease refer to\n`LICENSE <https://github.com/lisphilar/covid19-sir/blob/master/LICENSE>`__\nfile.\n\nCitation\n--------\n\nWe have no original papers the author and contributors wrote, but please\ncite this library as follows with version number\n(``import covsirphy as cs; cs.__version__``).\n\nCovsirPhy Development Team (2020-2022), CovsirPhy version [version\nnumber]: Python library for COVID-19 analysis with phase-dependent\nSIR-derived ODE models, https://github.com/lisphilar/covid19-sir\n\nIf you want to use SIR-F model, S-R trend analysis, phase-dependent\napproach to SIR-derived models, and other scientific method performed\nwith CovsirPhy, please cite the next Kaggle notebook.\n\nHirokazu Takaya (2020-2022), Kaggle Notebook, COVID-19 data with SIR\nmodel, https://www.kaggle.com/lisphilar/covid-19-data-with-sir-model\n\nWe can check the citation with the following script.\n\n.. code:: python\n\n    import covsirphy as cs\n    cs.__citation__\n\n.. |PyPI version| image:: https://badge.fury.io/py/covsirphy.svg\n   :target: https://badge.fury.io/py/covsirphy\n.. |Downloads| image:: https://pepy.tech/badge/covsirphy\n   :target: https://pepy.tech/project/covsirphy\n.. |PyPI - Python Version| image:: https://img.shields.io/pypi/pyversions/covsirphy\n   :target: https://badge.fury.io/py/covsirphy\n.. |Build Status| image:: https://semaphoreci.com/api/v1/lisphilar/covid19-sir/branches/master/shields_badge.svg\n   :target: https://semaphoreci.com/lisphilar/covid19-sir\n.. |GitHub license| image:: https://img.shields.io/github/license/lisphilar/covid19-sir\n   :target: https://github.com/lisphilar/covid19-sir/blob/master/LICENSE\n.. |Maintainability| image:: https://api.codeclimate.com/v1/badges/eb97eaf9804f436062b9/maintainability\n   :target: https://codeclimate.com/github/lisphilar/covid19-sir/maintainability\n.. |Test Coverage| image:: https://api.codeclimate.com/v1/badges/eb97eaf9804f436062b9/test_coverage\n   :target: https://codeclimate.com/github/lisphilar/covid19-sir/test_coverage\n.. |Open Source Helpers| image:: https://www.codetriage.com/lisphilar/covid19-sir/badges/users.svg\n   :target: https://www.codetriage.com/lisphilar/covid19-sir\n.. |Become a backer| image:: https://opencollective.com/covsirphy/tiers/backer.svg?avatarHeight=36&width=600\n   :target: https://opencollective.com/covsirphy\n',
    'author': 'Hirokazu Takaya',
    'author_email': 'lisphilar@outlook.jp',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lisphilar/covid19-sir',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.error import deprecate
from covsirphy.ode.mbase import ModelBase


class SIR(ModelBase):
    """
    SIR model.

    Args:
        population (int): total population
        rho (float)
        sigma (float)
    """
    # Model name
    NAME = "SIR"
    # names of parameters
    PARAMETERS = ["rho", "sigma"]
    DAY_PARAMETERS = ["1/beta [day]", "1/gamma [day]"]
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = {
        "x": ModelBase.S,
        "y": ModelBase.CI,
        "z": ModelBase.FR
    }
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array([1, 1, 1])
    # Variables that increases monotonically
    VARS_INCLEASE = [ModelBase.FR]
    # Example set of parameters and initial values
    EXAMPLE = {
        ModelBase.STEP_N: 180,
        ModelBase.N.lower(): 1_000_000,
        ModelBase.PARAM_DICT: {
            "rho": 0.2, "sigma": 0.075,
        },
        ModelBase.Y0_DICT: {
            ModelBase.S: 999_000, ModelBase.CI: 1000, ModelBase.FR: 0,
        },
    }

    def __init__(self, population, rho, sigma):
        # Total population
        self.population = self._ensure_population(population)
        # Non-dim parameters
        self.rho = rho
        self.sigma = sigma
        self.non_param_dict = {"rho": rho, "sigma": sigma}

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.

        Args:
            t (int): time steps
            X (numpy.array): values of th model variables

        Returns:
            (np.array)
        """
        n = self.population
        s, i, *_ = X
        dsdt = 0 - self.rho * s * i / n
        drdt = self.sigma * i
        didt = 0 - dsdt - drdt
        return np.array([dsdt, didt, drdt])

    @classmethod
    @deprecate(".param_range()", new=".guess()", version="2.19.1-zeta-fu1")
    def param_range(cls, taufree_df, population, quantiles=(0.1, 0.9)):
        """
        Deprecated. Define the value range of ODE parameters using (X, dX/dt) points.
        In SIR model, X is S, I and R here.

        Args:
            taufree_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - t (int): time steps (tau-free)
                    - columns with dimensional variables
            population (int): total population
            quantiles (tuple(int, int)): quantiles to cut, like confidence interval

        Returns:
            dict(str, tuple(float, float)): minimum/maximum values
        """
        cls._ensure_dataframe(taufree_df, name="taufree_df", columns=[cls.TS, *cls.VARIABLES])
        df = taufree_df.copy()
        df = df.loc[(df[cls.S] > 0) & (df[cls.CI] > 0)]
        n, t, s, i, r = population, df[cls.TS], df[cls.S], df[cls.CI], df[cls.FR]
        # rho = - n * (dS/dt) / S / I
        rho_series = 0 - n * s.diff() / t.diff() / s / i
        # sigma = (dR/dt) / I
        sigma_series = r.diff() / t.diff() / i
        # Calculate quantile
        _dict = {
            k: tuple(v.quantile(quantiles).clip(0, 1))
            for (k, v) in zip(["rho", "sigma"], [rho_series, sigma_series])
        }
        return _dict

    @classmethod
    @deprecate(".specialize()", new=".convert()", version="2.19.1-zeta-fu1")
    def specialize(cls, data_df, population):
        """
        Deprecated. Specialize the dataset for this model.

        Args:
            data_df (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - any columns
            population (int): total population in the place

        Returns:
            (pandas.DataFrame):
                    Index
                        reset index
                    Columns
                        - any columns @data_df has
                        - Susceptible (int): the number of susceptible cases
                        - Fatal or Recovered (int): total number of fatal/recovered cases
        """
        cls._ensure_dataframe(data_df, name="data_df", columns=cls.VALUE_COLUMNS)
        df = data_df.copy()
        # Calculate dimensional variables
        df[cls.S] = population - df[cls.C]
        df[cls.FR] = df[cls.F] + df[cls.R]
        return df

    @classmethod
    @deprecate(".restore()", new=".convert_reverse()", version="2.19.1-zeta-fu1")
    def restore(cls, specialized_df):
        """
        Deprecated. Restore Confirmed/Infected/Recovered/Fatal.
         using a dataframe with the variables of the model.

        Args:
            specialized_df (pandas.DataFrame): dataframe with the variables

                Index
                    reset index
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
                    - any columns

        Returns:
            (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - the other columns @specialzed_df has
        """
        df = specialized_df.copy()
        other_cols = list(set(df.columns) - set(cls.VALUE_COLUMNS))
        df[cls.C] = df[cls.CI] + df[cls.FR]
        df[cls.F] = 0
        df[cls.R] = df[cls.FR]
        return df.loc[:, [*cls.VALUE_COLUMNS, *other_cols]]

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.

        Returns:
            float
        """
        try:
            rt = self.rho / self.sigma
        except ZeroDivisionError:
            return None
        return round(rt, 2)

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        try:
            return {
                "1/beta [day]": int(tau / 24 / 60 / self.rho),
                "1/gamma [day]": int(tau / 24 / 60 / self.sigma)
            }
        except ZeroDivisionError:
            return {p: None for p in self.DAY_PARAMETERS}

    @classmethod
    def convert(cls, data, tau):
        """
        Divide dates by tau value [min] and convert variables to model-specialized variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min] or None (skip division by tau values)

        Returns:
            pandas.DataFrame:
                Index
                    - Date (pd.Timestamp): Observation date (available when @tau is None)
                    - t (int): time steps (available when @tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
        """
        # Convert to tau-free if tau was specified
        df = cls._convert(data, tau)
        # Conversion of variables
        df[cls.FR] = df[cls.F] + df[cls.R]
        return df.loc[:, [cls.S, cls.CI, cls.FR]]

    @classmethod
    def convert_reverse(cls, converted_df, start, tau):
        """
        Calculate date with tau and start date, and restore Susceptible/Infected/"Fatal or Recovered".

        Args:
            converted_df (pandas.DataFrame):
                Index
                    t: Dates divided by tau value (time steps)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
            start (pd.Timestamp): start date of simulation, like 14Apr2021
            tau (int): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        # Calculate date with tau and start date
        df = cls._convert_reverse(converted_df, start, tau)
        # Conversion of variables
        df[cls.F] = 0
        df[cls.R] = df.loc[:, cls.FR]
        return df.loc[:, [cls.DATE, cls.S, cls.CI, cls.F, cls.R]]

    @classmethod
    def guess(cls, data, tau, q=0.5):
        """
        With (X, dX/dt) for X=S, I, R, guess parameter values.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min]
            q (float or tuple(float,)): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict(str, float or pandas.Series): guessed parameter values with the quantile(s)

        Note:
            We can guess parameter values with difference equations as follows.
            - rho = - n * (dS/dt) / S / I
            - sigma = (dR/dt) / I
        """
        # Convert to tau-free and model-specialized dataset
        df = cls.convert(data=data, tau=tau)
        # Remove negative values and set variables
        df = df.loc[(df[cls.S] > 0) & (df[cls.CI] > 0)]
        n = df.loc[df.index[0], [cls.S, cls.CI, cls.FR]].sum()
        # Calculate parameter values with difference equation and tau-free data
        rho_series = 0 - n * df[cls.S].diff() / df[cls.S] / df[cls.CI]
        sigma_series = df[cls.FR].diff() / df[cls.CI]
        # Guess representative values
        return {
            "rho": cls._clip(rho_series.quantile(q=q), 0, 1),
            "sigma": cls._clip(sigma_series.quantile(q=q), 0, 1),
        }
